#!/usr/bin/env python
# id3v2rename copyright (c) 2002 by Charl P. Botha
# top-secret mechanisms, machinations and usability testing by Paul W. de Bruin
# $Id: id3v2rename.py,v 1.11 2004/07/30 13:39:28 cpbotha Exp $
# requires python > 2.0 and id3v2

# this code is under the GPL version 2 or later

import getopt
import os
import re
import shutil
import string
import sys

ID3V2='id3v2'
DEFAULT_RENAME_PATTERN='%a/%t/%n-%s.mp3'

def _disp_welcome():
    print("\n%s copyright (c) 2002 by Charl P. Botha http://cpbotha.net/\n"
          "Contributions/testing by Paul W. de Bruin http://pwdebruin.net/\n" %
          os.path.basename(sys.argv[0]))

def _disp_usage():
    sys.stderr.write("%s [options] file1 file2 ... fileN\n" % sys.argv[0])
    
    sys.stderr.write("""Options are:
    --help OR -h                                  : display this message
    --test OR -t                                  : do not actually move files
    --pattern=\"some_pattern\" OR -p 'some_pattern' : set rename pattern
    --numberpad=number OR -n number               : pad tracknumber with zeroes
                                                    up to 'number' (default 2)
    --empty-tag-string=\"some_string\"              : what to use if a tag value
                                                    can't be extracted
                                                    (default \"Unknown\")
    --empty-number-string=\"some_string\"           : what to use if a track num
                                                    can't be extracted
                                                    (default \"N\")
    --no-convert                                  : do not try to convert id3v1
                                                    only MP3s to id3v1/v2
    --copy                                        : do not rename/move MP3s,
                                                    rather copy them

    The rename pattern consists of a string containing tokens representing
    artist (\%a), album title (\%t), track number (\%n) and song title (\%s).
    The variables represented by these tokens will be extracted from the ID3v2
    tags of the MP3 that is being moved.  For example, you could do:
    id3v2rename.py --pattern=\"/mp3/new/%a/%t/%n-%s.mp3\" /incoming/*.mp3
    This would move all mp3s in /incoming to a directory hierarchy in /mp3.
    Directories that don't exist yet will be created.  If a file can't be
    simply renamed (because the destination is on a different disc/filesystem
    e.g.) a copy and unlink of the source will be attempted.
    """)

    sys.stderr.write("\n")


def _handle_filename(filename, pattern, test,
                     numberpad, empty_tag_string, empty_number_string,
                     convert, mp3copy):

    print("    %s" % (filename))

    # first check if the file exists; id3v2 can be quite stupid about this
    if not os.path.exists(filename):
        sys.stderr.write('*E* File does not exist.\n')
        return 0

    ##########################################################################
    # execute id3v2 --list, get output from pipe
    try:
        p = os.popen('%s --list \"%s\"' % (ID3V2, filename))
    except Exception,e:
        sys.stderr.write('*E* could not open.\n')
        return

    id3res = p.read()
    p.close()

    # make SURE we have some valid output
    VALIDV2_REGEXP = "id3v2 tag info for "
    VALIDV1_REGEXP = "id3v1 tag info for "
    valid_mo = re.search(VALIDV2_REGEXP, id3res)

    # if we don't, we can try converting
    if not valid_mo:

        if test:
            sys.stderr.write('*E* Could not extract ID3V2 tags.\n')
            sys.stderr.write('*I* Would have tried to convert from ID3V1, '
                             'but --test is active.\n')
            return 0
        
        elif convert and re.search(VALIDV1_REGEXP, id3res):
            # we should and can probably convert
            
            try:
                # we could not get id3v2 tags, let's try and convert to id3v1
                p2 = os.popen('%s --convert \"%s\"' % (ID3V2, filename))
                p2.close()
                # reopen the converted file
                p3 = os.popen('%s --list \"%s\"' % (ID3V2, filename))
                # get its tags out
                id3res = p3.read()
                # take care of the pipe
                p3.close()
            except Exception, e:
                sys.stderr.write('*E* Could not convert ID3V2->ID3V1.\n')
                return 0
            else:
                valid_mo = re.search(VALIDV2_REGEXP, id3res)
                if not valid_mo:
                    sys.stderr.write(
                        '*E* Could not extract ID3V2 tags after convert.\n')
                    return 0

        else:
            sys.stderr.write('*E* Could not extract ID3V2 tags.\n')
            return 0

    ##########################################################################

    # we will store our extracted tokens in a dict for easy substitution later
    tokens = {}
    # this translation table must sanitise the separate components
    # we remove /, \ and whatever else is going to confuse file ops
    ttable1 = string.maketrans('/\\', '__')
    # we also use lstrip()/rstrip() to throw away leading/trailing whitespace
    
    # now actually get out all the tokens
    # \)? means zero or one )... this is to compensate for changes in
    # id3v2 output
    title_mo = re.search("Title/songname/content description\)?:\s*(.*)", id3res)
    if title_mo:
        tokens['s'] = title_mo.group(1).lstrip().rstrip().translate(ttable1)
    else:
        tokens['s'] = empty_tag_string

    artist_mo = re.search("Lead performer\(s\)/Soloist\(s\)\)?:\s*(.*)", id3res)
    if artist_mo:
        tokens['a'] = artist_mo.group(1).lstrip().rstrip().translate(ttable1)
    else:
        tokens['a'] = empty_tag_string

    album_mo = re.search("Album/Movie/Show title\)?:\s*(.*)", id3res)
    if album_mo:
        tokens['t'] = album_mo.group(1).lstrip().rstrip().translate(ttable1)
    else:
        tokens['t'] = empty_tag_string

    track_mo = re.search("Track number/Position in set\)?:\s*(.*)", id3res)
    if track_mo:
        tokens['n'] = track_mo.group(1).lstrip().rstrip().translate(ttable1)
    else:
        tokens['n'] = empty_number_string

    # now pad the track number
    # first create the format string itself
    fstr = "%0" + str(numberpad) + "d"
    try:
        # first sanitise the tracknum
        tokens['n'] = re.search('([0-9]+).*', tokens['n']).group(1)
        tokens['n'] = fstr % int(tokens['n'])
    except Exception,e:
        sys.stderr.write('*W* Could not pad tracknum, trying to extract from filename.\n')
        basename = os.path.basename(filename)
        fmo = re.search('^([0-9]+)',basename)

        if fmo:
            tempTrackNum = int(fmo.group(1))
            tokens['n'] = fstr % tempTrackNum

            if not test:
                # write this track number into the id3v2 information
                # only if we're NOT in testing mode
                writeTrackCommand = '%s --track \"%d\" \"%s\"' % (ID3V2, tempTrackNum, filename)
                p4 = os.popen(writeTrackCommand)
                p4.close()
            
        else:
            sys.stderr.write('*E* Could not deduce track number.\n')
            return 0

    ##########################################################################
    # we have all the tokens, so let's construct the filename according to
    # the pattern
    new_filename = ''
    prev = 'a'
    for c in pattern:
        if c == '%':
            # just gobble these up
            pass
        elif prev == '%':
            # do stuff based on token
            new_filename += tokens[c]
        else:
            # this is not part of a token, just append to new_filename
            new_filename += c

        # store as previous character for next iteration
        prev = c

    # once again we sanitise the full filename by removing all spaces
    ttable2 = string.maketrans(" ", "_")
    new_filename = new_filename.translate(ttable2)
    # also have os.path work its magic
    new_filename = os.path.normpath(new_filename)

    ##########################################################################
    # this is what we're going to do
    if mp3copy:
        print "<-> %s" % (new_filename)
    else:
        print "--> %s" % (new_filename)


    if not test:
        try:
            # first find out if directories need to be created
            dn = os.path.dirname(new_filename)
            if len(dn) > 0 and not os.path.exists(dn):
                # and create them if required
                os.makedirs(dn)
            if mp3copy:
                # copy, don't move
                shutil.copy(filename, new_filename)
            else:
		try:
                    # now we can finally rename the f*cker
                    os.rename(filename, new_filename)
                except Exception, e:
                    # this means the rename didn't work, which could be because
                    # we're trying to rename across filesystems; so let's try
                    # copy and delete
                    # we're now in the outer try clause, so that should catch
                    # any errors that the copy and unlink cause
                    sys.stderr.write(
                        '*I* Could not rename, trying copy/unlink.\n')
                    shutil.copy(filename, new_filename)
                    os.unlink(filename)
            
        except Exception, e:
            sys.stderr.write('*E* Problem moving file: %s.\n' % (e))
            return 0
        else:
            return 1

    return 0

def main():
    _disp_welcome()

    try:
        # 'p:' means -p with something after
        optlist, args = getopt.getopt(
            sys.argv[1:], 'cn:p:th',
            ['empty-tag-string=', 'empty-number-string=', 'copy',
             'no-convert', 'numberpad=', 'pattern=', 'test', 'help'])
    except getopt.GetoptError,e:
        _disp_usage()
        sys.exit(1)

    # set all defaults here before beginning
    pattern = DEFAULT_RENAME_PATTERN
    numberpad = 2
    test = 0
    convert = 1
    mp3copy = 0
    empty_tag_string = "Unknown"
    empty_number_string = "N"

    for o, a in optlist:
        if o in ('--empty-tag-string',):
            empty_tag_string = a

        elif o in ('--empty-number-string',):
            empty_number_string = a
            
        elif o in ('-c', '--copy'):
            mp3copy = 1
            
        elif o in ('--no-convert',):
            convert = 0
            
        elif o in ('-n', '--numberpad'):
            try:
                numberpad = int(a)
            except:
                sys.stderr.write(
                    'Invalid numberpad specified.  Defaulting to 2.\n')
                
        elif o in ('-p', '--pattern') and len(a) > 0:
            pattern = a
            
        elif o in ('-t', '--test'):
            test = 1
            
        elif o in ('-h', '--help'):
            _disp_usage()
            sys.exit(0)

    if len(args) == 0:
        _disp_usage()
        sys.exit(1)
        
    files_moved = 0
    for filename in args:
        files_moved += _handle_filename(filename, pattern, test,
                                        numberpad, empty_tag_string,
                                        empty_number_string, convert, mp3copy)

    action = 'moved'
    if mp3copy:
        action = 'copied'
        
    print('\n%d / %d files %s.' % (files_moved, len(args), action))
        
if __name__ == '__main__':
    main()
